from typing import Optional

from indi.device import Driver, properties
from indi.message import GetProperties
from indi.transport.client.tcp import TCP as TCPClient


class Proxy(Driver):
    address: Optional[str] = None
    port = 7624

    general = properties.Group(
        "GENERAL",
        vectors=dict(connection=properties.Standard("CONNECTION", onchange="connect"),),
    )

    def __init__(self, *args, **kwargs):
        self._connection = None
        self._client = TCPClient(self.address, self.port)
        super().__init__(*args, **kwargs)

    def message_from_client(self, message):
        if getattr(message, "device", None) in (None, self.name):
            super().message_from_client(message)

        if self._connection:
            self._connection.send_message(message)

    def _message_from_server(self, message):
        self._router.process_message(message, self)

    def accepts(self, device):
        return True

    def connect(self, sender):
        connected = self.get_group("general").connection.connect.bool_value
        if connected:
            self._connection = self._client.connect(self._message_from_server)
            self._connection.send_message(GetProperties(version="1.0"))
        else:
            if self._connection:
                self._connection.close()
            self._connection = None

    @property
    def client(self):
        return self._client



class Permissions:
    READ_ONLY = "ro"
    WRITE_ONLY = "wo"
    READ_WRITE = "rw"


class State:
    IDLE = "Idle"
    OK = "Ok"
    BUSY = "Busy"
    ALERT = "Alert"


class SwitchRule:
    ONE_OF_MANY = "OneOfMany"
    AT_MOST_ONE = "AtMostOne"
    ANY_OF_MANY = "AnyOfMany"


class SwitchState:
    ON = "On"
    OFF = "Off"


class BLOBEnable:
    NEVER = "Never"
    ALSO = "Also"
    ONLY = "Only"



import logging
from functools import wraps

from .driver import Driver
from .proxy import Proxy


def non_blocking(fun):
    @wraps(fun)
    def wrapped(*args, **kwargs):
        import threading

        def inner_fun():
            try:
                fun(*args, **kwargs)
            except Exception as e:
                logging.exception(e)

        th = threading.Thread(target=inner_fun)
        th.start()

    return wrapped



import base64
import logging
from typing import Type, Union

from indi.device.properties.instance.vectors import Vector
from indi.message import checks, const, parts


class Element:
    def_message_class: Union[Type[parts.DefSwitch], Type[parts.DefNumber], Type[parts.DefLight], Type[parts.DefBLOB], Type[parts.DefText]]
    set_message_class: Union[Type[parts.OneSwitch], Type[parts.OneNumber], Type[parts.OneLight], Type[parts.OneBLOB], Type[parts.OneText]]

    def __init__(self, vector: Vector, definition):
        self._vector = vector
        self._definition = definition
        self._value = definition.default
        self._enabled = definition.enabled

    @property
    def name(self):
        return self._definition.name

    @property
    def device(self):
        return self._vector.device

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        self._enabled = value

    @property
    def value(self):
        self.device.trigger_callback(self._definition.onread, self)
        return self._value

    @value.setter
    def value(self, value):
        logging.debug(
            f"Element: setting value of element {self._definition.name} to {value}"
        )
        prev_value = self._value
        self._value = self.check_value(value)
        self.device.send_message(self._vector.to_set_message())

        if prev_value != self._value:
            self.device.trigger_callback(self._definition.onchange, self)
            self.device.trigger_callback(self._vector.onchange, self._vector)
            self.device.trigger_callback(
                self._vector.group.onchange, self._vector.group
            )
            self.device.trigger_callback(self.device.onchange, self.device)

    def set_value(self, value):
        if self._definition.onwrite is not None:
            self.device.trigger_callback(self._definition.onwrite, self, value=value)
        else:
            self.value = value

    def reset_value(self, value):
        self._value = self.check_value(value)

    def check_value(self, value):
        return value

    def to_def_message(self):
        return self.def_message_class(
            name=self._definition.name, value=self.value, label=self._definition.label,
        )

    def to_set_message(self):
        return self.set_message_class(name=self._definition.name, value=self.value)


class Number(Element):
    def_message_class = parts.DefNumber
    set_message_class = parts.OneNumber

    def to_def_message(self):
        return self.def_message_class(
            name=self._definition.name,
            value=self.value,
            label=self._definition.label,
            format=self._definition.format,
            min=self._definition.min,
            max=self._definition.max,
            step=self._definition.step,
        )


class Text(Element):
    def_message_class = parts.DefText
    set_message_class = parts.OneText


class Switch(Element):
    def_message_class = parts.DefSwitch
    set_message_class = parts.OneSwitch

    def check_value(self, value):
        value = checks.dictionary(value, const.SwitchState)
        return self._vector.apply_rule(self, value)

    @property
    def bool_value(self):
        return self.value == const.SwitchState.ON

    @bool_value.setter
    def bool_value(self, value):
        self.value = const.SwitchState.ON if value else const.SwitchState.OFF

    def reset_bool_value(self, value):
        self._value = const.SwitchState.ON if value else const.SwitchState.OFF


class Light(Element):
    def_message_class = parts.DefLight
    set_message_class = parts.OneLight

    def check_value(self, value):
        return checks.dictionary(value, const.State)


class BLOB(Element):
    def_message_class = parts.DefBLOB
    set_message_class = parts.OneBLOB

    @property
    def bin_value(self):
        return base64.b64decode(self.value)

    @bin_value.setter
    def bin_value(self, value):
        self.value = base64.b64encode(value)



from typing import Type, Union

from indi.message import checks, const
from indi.message.IndiMessage import IndiMessage
from indi.message.parts import OneBLOB, OneLight, OneNumber, OneSwitch, OneText


class SetVector(IndiMessage):
    child_class: Union[Type[OneBLOB], Type[OneLight], Type[OneNumber], Type[OneSwitch], Type[OneText]]
    from_device = True

    def __init__(
        self,
        device,
        name,
        state,
        timeout=None,
        timestamp=None,
        message=None,
        children=None,
        **junk
    ):
        self.device = device
        self.name = name
        self.state = checks.dictionary(state, const.State)
        self.timeout = timeout
        self.timestamp = timestamp
        self.message = message
        self.children = checks.children(children, self.child_class)


class SetBLOBVector(SetVector):
    child_class = OneBLOB


class SetLightVector(SetVector):
    child_class = OneLight


class SetNumberVector(SetVector):
    child_class = OneNumber


class SetSwitchVector(SetVector):
    child_class = OneSwitch


class SetTextVector(SetVector):
    child_class = OneText



import logging

from indi.message import IndiMessage
from indi.routing import Client
from websocket_server import WebsocketServer

logger = logging.getLogger(__name__)


class ConnectionHandler(Client):
    def __init__(self, server, client, router):
        self.server = server
        self.client = client
        self.router = router

    def message_from_device(self, message):
        data = message.to_string()
        try:
            self.server.send_message(self.client, data)
        except:
            pass

    def message_from_client(self, message):
        if self.router:
            self.router.process_message(message, sender=self)

    def text_message_from_client(self, text_message):
        message = IndiMessage.from_string(text_message)
        self.message_from_client(message)


class WebSocket:
    def __init__(self, address="0.0.0.0", port=8001, router=None):
        self.address = address
        self.port = port
        self.router = router
        self.clients = {}

    def start(self):
        logger.info("Starting WebSocket INDI server on %s:%s", self.address, self.port)
        server = WebsocketServer(self.port, host=self.address)
        server.set_fn_new_client(self._new_client)
        server.set_fn_client_left(self._client_left)
        server.set_fn_message_received(self._message_received)
        server.run_forever()

    def _new_client(self, client, server):
        handler = ConnectionHandler(server, client, self.router)
        self.clients[client["id"]] = handler
        self.router.register_client(handler)

    def _client_left(self, client, server):
        if client:
            handler = self.clients[client["id"]]
            del self.clients[client["id"]]
            self.router.unregister_client(handler)

    def _message_received(self, client, server, message):
        self.clients[client["id"]].text_message_from_client(message)



from indi.message.IndiMessage import IndiMessage


class DelProperty(IndiMessage):
    from_device = True

    def __init__(self, device, name=None, timestamp=None, message=None, **junk):
        self.device = device
        self.name = name
        self.timestamp = timestamp
        self.message = message



from typing import Dict, List, Type

from indi.device import Driver
from indi.routing import Router


class DevicePool:
    device_classes: List[type] = []

    devices: Dict[str, Driver] = {}

    @classmethod
    def register(cls, device_class: Type[Driver]):
        cls.device_classes.append(device_class)

    @classmethod
    def init(cls, router: Router):
        for device_class in cls.device_classes:
            dev = device_class(router=router)
            cls.devices[dev.name] = dev

    @classmethod
    def get(cls, name: str) -> Driver:
        return cls.devices[name]



from typing import Any, Optional, Type, Union

from indi.device.properties.instance import elements as instance_elements
from indi.message import const
from indi.typing import CallbackDefinition


class Element:
    instance_class = Union[Type[instance_elements.BLOB], Type[instance_elements.Light], Type[instance_elements.Number], Type[instance_elements.Switch], Type[instance_elements.Text]]
    default_value: Any

    def __init__(
        self,
        name: str,
        label: Optional[str] = None,
        default=None,
        enabled: bool = True,
        onchange: Optional[CallbackDefinition] = None,
        onwrite: Optional[CallbackDefinition] = None,
        onread: Optional[CallbackDefinition] = None,
    ):
        self.name = name
        self.label = label or name
        self.default = self.default_value if default is None else default
        self.enabled = enabled

        self.onchange = onchange
        self.onwrite = onwrite
        self.onread = onread

    def instance(self, vector):
        return self.instance_class(vector, self)


class Number(Element):
    instance_class = instance_elements.Number
    default_value = 0.0

    def __init__(self, *args, format="%f", min=0, max=None, step=0, **kwargs):
        super().__init__(*args, **kwargs)
        self.format = format
        self.min = min
        self.max = max if max else min
        self.step = step


class Text(Element):
    instance_class = instance_elements.Text
    default_value = ""


class Switch(Element):
    instance_class = instance_elements.Switch
    default_value = const.SwitchState.OFF


class Light(Element):
    instance_class = instance_elements.Light
    default_value = const.State.OK


class BLOB(Element):
    instance_class = instance_elements.BLOB
    default_value = ""



from indi.message import checks, const
from indi.message.IndiMessage import IndiMessage


class EnableBLOB(IndiMessage):
    from_client = True

    def __init__(self, device, value, name=None, **junk):
        self.device = device
        self.name = name
        self.value = checks.dictionary(value, const.BLOBEnable)



import logging
import socket
import threading
import sys

from indi.routing import Client
from indi.transport import Buffer

logger = logging.getLogger(__name__)


class ConnectionHandler(Client):
    def __init__(self, router, stdin, stdout):
        self.stdin = stdin
        self.stdout = stdout
        self.buffer = Buffer()
        self.router = router
        self.sender_lock = threading.Lock()
        if self.router:
            self.router.register_client(self)

    def handle(self):
        try:
            self.wait_for_messages()
        except:
            logger.exception("Uncaugth error")

        logger.info("Stopping INDIpy server on TTY")
        self.close()

    def wait_for_messages(self):
        while True:
            logger.debug(f"Waiting for data")
            message = self._read()
            if not message:
                logger.debug(f"No data, exiting")
                break
            logger.debug("Received data: %s", message)
            self.buffer.append(message)
            self.buffer.process(self.message_from_client)

    def message_from_client(self, message):
        if self.router:
            self.router.process_message(message, sender=self)

    def message_from_device(self, message):
        data = message.to_string().decode("latin1")
        with self.sender_lock:
            logger.debug("Sending data: %s", data)
            self._write(data)

    def close(self):
        if self.router:
            self.router.unregister_client(self)

    def _read(self):
        return self.stdin.readline()

    def _write(self, data):
        self.stdout.write(data)
        self.stdout.flush()


class TTY:
    def __init__(self, router=None, stdin=None, stdout=None):
        self.router = router
        self.stdin = stdin or sys.stdin
        self.stdout = stdout or sys.stdout

    def start(self):
        logger.info("Starting INDIpy server on TTY")
        conn_handler = ConnectionHandler(self.router, self.stdin, self.stdout)
        conn_handler.handle()



from typing import Type, Union

from indi.message import checks, const
from indi.message.IndiMessage import IndiMessage
from indi.message.parts import DefBLOB, DefLight, DefNumber, DefSwitch, DefText


class DefVector(IndiMessage):
    from_device = True
    children_class: Union[Type[DefBLOB], Type[DefLight], Type[DefNumber], Type[DefSwitch], Type[DefText]]

    def __init__(
        self,
        device,
        name,
        state,
        label=None,
        group=None,
        timestamp=None,
        message=None,
        children=None,
        **junk
    ):
        self.device = device
        self.name = name
        self.state = checks.dictionary(state, const.State)
        self.label = label
        self.group = group
        self.timestamp = timestamp
        self.message = message
        self.children = checks.children(children, self.children_class)


class DefWritableVector(DefVector):
    def __init__(self, *args, perm, timeout=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.perm = checks.dictionary(perm, const.Permissions)
        self.timeout = timeout


class DefBLOBVector(DefWritableVector):
    children_class = DefBLOB


class DefLightVector(DefVector):
    children_class = DefLight


class DefNumberVector(DefWritableVector):
    children_class = DefNumber


class DefSwitchVector(DefWritableVector):
    children_class = DefSwitch

    def __init__(self, *args, rule, **kwargs):
        super().__init__(*args, **kwargs)
        self.rule = checks.dictionary(rule, const.SwitchRule)


class DefTextVector(DefWritableVector):
    children_class = DefText



from typing import Type, Union

from indi import message
from indi.message import checks, const


class Vector:
    def_message_class: Type[message.DefVector]
    set_message_class: Union[Type[message.SetBLOBVector], Type[message.SetLightVector], Type[message.SetNumberVector], Type[message.SetSwitchVector], Type[message.SetTextVector]]
    new_message_class: Union[Type[message.NewBLOBVector], Type[message.NewNumberVector], Type[message.NewSwitchVector], Type[message.NewTextVector]]

    def __init__(self, group, definition):
        self._state = definition.state
        self._enabled = definition.enabled
        self._group = group
        self._definition = definition
        self._elements = {k: v.instance(self) for k, v in definition.elements.items()}
        self._elements_by_name = {v.name: v for k, v in self._elements.items()}

    def __getattr__(self, item):
        element = self._elements.get(item)
        if element:
            return element

        return self.__getattribute__(item)

    @property
    def device(self):
        return self._group.device

    @property
    def group(self):
        return self._group

    @property
    def name(self):
        return self._definition.name

    @property
    def onchange(self):
        return self._definition.onchange

    @property
    def enabled(self):
        return self._enabled and self.group.enabled

    @enabled.setter
    def enabled(self, value):
        self._enabled = value
        self.device.send_message(self.to_def_message())
        self.device.send_message(self.to_set_message())

    @property
    def state_(self):
        return self._state

    @state_.setter
    def state_(self, value):
        self._state = checks.dictionary(value, const.State)
        self.device.send_message(self.to_set_message())

    def to_def_message(self):
        if not self.enabled:
            return message.DelProperty(
                device=self.device.name,
                name=self._definition.name,
                timestamp=message.now(),
            )

        elements = [e.to_def_message() for k, e in self._elements.items() if e.enabled]
        return self.def_message_class(
            device=self.device.name,
            name=self._definition.name,
            group=self._group.name,
            label=self._definition.label,
            state=self._state,
            perm=self._definition.perm,
            timeout=self._definition.timeout,
            timestamp=message.now(),
            children=elements,
        )

    def to_set_message(self):
        if not self.enabled:
            return None

        elements = [e.to_set_message() for k, e in self._elements.items() if e.enabled]
        return self.set_message_class(
            device=self.device.name,
            name=self._definition.name,
            state=self._state,
            timeout=self._definition.timeout,
            timestamp=message.now(),
            children=elements,
        )

    def from_new_message(self, msg):
        for child in msg.children:
            self._elements_by_name[child.name].set_value(child.value)


class NumberVector(Vector):
    def_message_class = message.DefNumberVector
    set_message_class = message.SetNumberVector
    new_message_class = message.NewNumberVector


class TextVector(Vector):
    def_message_class = message.DefTextVector
    set_message_class = message.SetTextVector
    new_message_class = message.NewTextVector


class SwitchVector(Vector):
    def_message_class = message.DefSwitchVector
    set_message_class = message.SetSwitchVector
    new_message_class = message.NewSwitchVector

    @property
    def selected_value(self):
        res = self.selected_values
        if len(res) > 1:
            raise Exception("Multiple values selected")

        try:
            return res[0]
        except IndexError:
            return None

    @selected_value.setter
    def selected_value(self, value):
        self.selected_values = [value]

    @property
    def selected_values(self):
        res = []
        for k, el in self._elements.items():
            if el.bool_value:
                res.append(el.name)
        return res

    @selected_values.setter
    def selected_values(self, values):
        for v in values:
            if v not in self._elements_by_name:
                raise Exception()

        for k, el in self._elements.items():
            new_value = el.name in values
            if el.bool_value != new_value:
                el.bool_value = new_value

    def reset_selected_value(self, value):
        self.reset_selected_values([value])

    def reset_selected_values(self, values):
        for v in values:
            if v not in self._elements_by_name:
                raise Exception()

        for k, el in self._elements.items():
            el.reset_bool_value(el.name in values)

    def apply_rule(self, sender, new_value):
        if new_value == const.SwitchState.ON:
            if self._definition.rule in (
                const.SwitchRule.AT_MOST_ONE,
                const.SwitchRule.ONE_OF_MANY,
            ):
                for k, el in self._elements.items():
                    if el != sender and el._value == const.SwitchState.ON:
                        el._value = const.SwitchState.OFF
        else:
            if self._definition.rule in (const.SwitchRule.ONE_OF_MANY,):
                if (
                    len(
                        [
                            el
                            for k, el in self._elements.items()
                            if el != sender and el._value == const.SwitchState.ON
                        ]
                    )
                    <= 0
                ):
                    new_value = const.SwitchState.ON

        return new_value

    def to_def_message(self):
        if not self.enabled:
            return message.DelProperty(
                device=self.device.name,
                name=self._definition.name,
                timestamp=message.now(),
            )

        elements = [e.to_def_message() for k, e in self._elements.items() if e.enabled]
        return self.def_message_class(
            device=self.device.name,
            name=self._definition.name,
            group=self._group.name,
            label=self._definition.label,
            state=self._state,
            perm=self._definition.perm,
            rule=self._definition.rule,
            timeout=self._definition.timeout,
            timestamp=message.now(),
            children=elements,
        )


class LightVector(Vector):
    def_message_class = message.DefLightVector
    set_message_class = message.SetLightVector

    def to_def_message(self):
        if not self.enabled:
            return message.DelProperty(
                device=self.device.name,
                name=self._definition.name,
                timestamp=message.now(),
            )

        elements = [e.to_def_message() for k, e in self._elements.items() if e.enabled]
        return self.def_message_class(
            device=self.device.name,
            name=self._definition.name,
            group=self.group.name,
            label=self._definition.label,
            state=self._state,
            timestamp=message.now(),
            children=elements,
        )

    def to_set_message(self):
        if not self.enabled:
            return None

        elements = [e.to_set_message() for k, e in self._elements.items() if e.enabled]
        return self.set_message_class(
            device=self.device.name,
            name=self._definition.name,
            state=self._state,
            timestamp=message.now(),
            children=elements,
        )


class BLOBVector(Vector):
    def_message_class = message.DefBLOBVector
    set_message_class = message.SetBLOBVector
    new_message_class = message.NewBLOBVector



import sys
import indi

from functools import reduce

from indi.device import properties
from indi.device.properties import const

"""
Standard property vector factory as defined in http://indilib.org/develop/developer-manual/101-standard-properties.html
"""

_STANDARD_VECTORS = {
    # GENERAL
    "CONNECTION": {
        "class": properties.SwitchVector,
        "elements": dict(connect="CONNECT", disconnect="DISCONNECT",),
        "kwargs": {"rule": const.SwitchRule.ONE_OF_MANY, "default_on": "DISCONNECT",},
    },
    "UPLOAD_MODE": {
        "class": properties.SwitchVector,
        "elements": dict(
            client="UPLOAD_CLIENT", local="UPLOAD_LOCAL", both="UPLOAD_BOTH",
        ),
        "kwargs": {"rule": const.SwitchRule.ONE_OF_MANY,},
    },
    "ACTIVE_DEVICES": {
        "class": properties.TextVector,
        "elements": {},
        "kwargs": {"perm": const.Permissions.READ_ONLY,},
    },
    # CCD
    "CCD_EXPOSURE": {
        "class": properties.NumberVector,
        "elements": dict(time="CCD_EXPOSURE_VALUE",),
    },
    "CCD_COMPRESSION": {
        "class": properties.SwitchVector,
        "elements": dict(compress="CCD_COMPRESS", raw="CCD_RAW",),
        "kwargs": {"rule": const.SwitchRule.ANY_OF_MANY,},
    },
    # FOCUSER
    "ABS_FOCUS_POSITION": {
        "class": properties.NumberVector,
        "elements": dict(position="FOCUS_ABSOLUTE_POSITION",),
    },
}


def Standard(name, **kwargs):
    definition = _STANDARD_VECTORS[name]

    def_kwargs = definition["kwargs"] if "kwargs" in definition else {}
    def_kwargs["elements"] = {
        k: definition["class"].element_class(n)
        for k, n in definition["elements"].items()
    }

    kwargs = {
        **def_kwargs,
        **kwargs,
    }

    return definition["class"](name, **kwargs)


def DriverInfo(name=None, exec=None, version=None, interface=None):
    name = name or "INDIpy Driver"
    exec = exec or sys.argv[0]
    version = version or indi.__version__
    interface = interface or []
    interface = reduce(
        lambda a, b: a | b,
        interface,
        const.DriverInterface.GENERAL
    )

    elements = {
        "name": properties.TextVector.element_class(name="DRIVER_NAME", default=name),
        "exec": properties.TextVector.element_class(name="DRIVER_EXEC", default=exec),
        "version": properties.TextVector.element_class(name="DRIVER_VERSION", default=version),
        "interface": properties.TextVector.element_class(name="DRIVER_INTERFACE", default=interface),
    }

    return properties.TextVector("DRIVER_INFO", elements=elements, perm=const.Permissions.READ_ONLY)



import xml.etree.ElementTree as ET

from indi.message import checks, const


class IndiMessagePart:
    def __init__(self, name, value, **junk):
        self.name = name
        self.value = self.check_value(value)

    def check_value(self, value):
        return value

    @classmethod
    def tag_name(cls):
        return cls.__name__[:1].lower() + cls.__name__[1:]

    @classmethod
    def _all_subclasses(cls):
        return cls.__subclasses__()

    @classmethod
    def from_xml(cls, xml):
        tag = xml.tag
        message_class = None

        for subclass in cls._all_subclasses():
            if subclass.tag_name() == tag:
                message_class = subclass

        if not message_class:
            raise Exception(f"Invalid part: {tag}")

        kwargs = xml.attrib

        kwargs["value"] = xml.text.strip() if xml.text else None

        return message_class(**kwargs)

    def to_xml(self, parent):
        kwargs = {
            k: str(v)
            for k, v in self.__dict__.items()
            if v is not None and k not in ("value",)
        }

        element = ET.SubElement(parent, self.__class__.tag_name(), **kwargs)
        if self.value is not None:
            element.text = str(self.value)

        return element


class DefIndiMessagePart(IndiMessagePart):
    def __init__(self, name, value=None, label=None, **junk):
        super().__init__(name=name, value=value)
        self.label = label


class DefBLOB(DefIndiMessagePart):
    pass


class DefLight(DefIndiMessagePart):
    pass


class DefNumber(DefIndiMessagePart):
    def __init__(self, name, format, min, max, step, value=None, label=None, **junk):
        super().__init__(name=name, value=value, label=label)
        self.format = format
        self.min = min
        self.max = max
        self.step = step


class DefSwitch(DefIndiMessagePart):
    pass


class DefText(DefIndiMessagePart):
    pass


class OneBLOB(IndiMessagePart):
    def __init__(self, name, size, format, value, **junk):
        super().__init__(name=name, value=value)
        self.size = size
        self.format = format


class OneLight(IndiMessagePart):
    def check_value(self, value):
        return checks.dictionary(value, const.State)


class OneNumber(IndiMessagePart):
    pass


class OneSwitch(IndiMessagePart):
    def check_value(self, value):
        return checks.dictionary(value, const.SwitchState)


class OneText(IndiMessagePart):
    pass



from .definition.elements import BLOB, Light, Number, Switch, Text
from .definition.group import Group
from .definition.vectors import (BLOBVector, LightVector, NumberVector,
                                 SwitchVector, TextVector)
from .definition.standard import Standard, DriverInfo



from indi.message.IndiMessage import IndiMessage


class OneLight(IndiMessage):
    from_device = True

    def __init__(self, name, value, **junk):
        self.name = name
        self.value = value



from typing import Type, Union

from indi import message
from indi.client import elements
from indi.message import checks


class Vector:
    def_message_class: Type[message.DefVector]
    set_message_class: Union[Type[message.SetBLOBVector], Type[message.SetLightVector], Type[message.SetNumberVector], Type[message.SetSwitchVector], Type[message.SetTextVector]]
    new_message_class: Union[Type[message.NewBLOBVector], Type[message.NewNumberVector], Type[message.NewSwitchVector], Type[message.NewTextVector]]
    children_class: Type[elements.Element]

    def __init__(self, device, msg: message.DefVector):
        self.group = msg.group
        self.name = msg.name
        self.label = msg.label
        self.timestamp = msg.timestamp
        self.message = msg.message
        self.state = msg.state
        self.device = device

        children = checks.children(
            [elements.Element.from_message(self, c) for c in msg.children],
            self.children_class,
        )
        self.elements = {ch.name: ch for ch in children}

    def __getitem__(self, key):
        return self.elements[key]

    def __contains__(self, key):
        return key in self.elements

    @classmethod
    def from_message(cls, device, msg):
        for subclass in cls.__subclasses__():
            if isinstance(msg, subclass.def_message_class):
                return subclass(device, msg)
        return None

    def get_element(self, name):
        return self.elements.get(name)

    def process_message(self, msg):
        if isinstance(msg, self.set_message_class):
            old_state = self.state
            self.state = msg.state

            if old_state != self.state:
                self.device.client.trigger_update(self, "state")

            for ch in msg.children:
                el = self.elements.get(ch.name)
                if el:
                    el.process_message(ch)

    def submit(self):
        ch = []
        for k, el in self.elements.items():
            ch.append(el.to_new_message())

        msg = self.new_message_class(
            device=self.device.name,
            name=self.name,
            timestamp=message.now(),
            children=ch,
        )

        self.device.send_message(msg)


class NumberVector(Vector):
    def_message_class = message.DefNumberVector
    set_message_class = message.SetNumberVector
    new_message_class = message.NewNumberVector
    children_class = elements.Number


class SwitchVector(Vector):
    def_message_class = message.DefSwitchVector
    set_message_class = message.SetSwitchVector
    new_message_class = message.NewSwitchVector
    children_class = elements.Switch


class TextVector(Vector):
    def_message_class = message.DefTextVector
    set_message_class = message.SetTextVector
    new_message_class = message.NewTextVector
    children_class = elements.Text


class BLOBVector(Vector):
    def_message_class = message.DefBLOBVector
    set_message_class = message.SetBLOBVector
    new_message_class = message.NewBLOBVector
    children_class = elements.BLOB


class LightVector(Vector):
    def_message_class = message.DefLightVector
    set_message_class = message.SetLightVector
    children_class = elements.Light



import logging

from indi.device import Driver
from indi.message import Message, now
from indi.routing import Router


class Handler(logging.Handler):
    def __init__(self, router: Router, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.router = router

    def emit(self, record):
        timestamp = record.timestamp if hasattr(record, "timestamp") else now()

        device = getattr(record, "device", None)
        if isinstance(device, Driver):
            device = device.name

        msg = Message(
            device=device, timestamp=timestamp, message=self.format(record)
        )

        self.router.process_message(message=msg)



import logging
import threading

from indi.message import EnableBLOB, NewBLOBVector, const

logger = logging.getLogger(__name__)


class Router:
    _instance = None

    DEFAULT_BLOB_POLICY = const.BLOBEnable.NEVER

    def __init__(self):
        self.clients = []
        self.devices = []
        self.blob_routing = {}

    @classmethod
    def instance(cls):
        if not cls._instance:
            cls._instance = cls()
        return cls._instance

    def register_device(self, device):
        self.devices.append(device)

    def register_client(self, client):
        logger.debug("Router: registering client %s", client)
        self.clients.append(client)
        self.blob_routing[client] = {}

    def unregister_client(self, client):
        logger.debug("Router: unregistering client %s", client)
        if client in self.clients:
            self.clients.remove(client)
        if client in self.blob_routing:
            del self.blob_routing[client]

    def process_message(self, message, sender=None):
        is_blob = isinstance(message, NewBLOBVector)

        if message.from_client:
            if isinstance(message, EnableBLOB):
                self.process_enable_blob(message, sender)

            for device in self.devices:
                if not device == sender and (
                    not message.device or device.accepts(message.device)
                ):
                    device.message_from_client(message)

        if message.from_device:
            for client in self.clients:
                if not client == sender:
                    device = getattr(message, "device", None)
                    client_blob_policy = self.blob_routing.get(client, {}).get(
                        device, self.DEFAULT_BLOB_POLICY
                    )
                    if (
                        is_blob
                        and client_blob_policy
                        in (const.BLOBEnable.ALSO, const.BLOBEnable.ONLY,)
                    ) or (not is_blob and client_blob_policy == const.BLOBEnable.NEVER):

                        def send():
                            client.message_from_device(message)

                        th = threading.Thread(target=send, daemon=True)
                        th.start()

    def process_enable_blob(self, message, sender):
        self.blob_routing[sender][message.name] = message.value



import logging
import socket
import threading

from indi.transport import Buffer


class ConnectionHandler:
    def __init__(self, client_socket, callback):
        self.buffer = Buffer()
        self.client_socket = client_socket
        self.callback = callback
        self.sender_lock = threading.Lock()

    def wait_for_messages(self):
        while True:
            logging.debug(f"TCP: waiting for data")
            message = self.client_socket.recv(1024)
            if not message:
                logging.debug(f"TCP: no data, breaking")
                break
            logging.debug(f"TCP: got data: {message}")
            self.buffer.append(message.decode("latin1"))
            self.buffer.process(self.message_from_server)

    def message_from_server(self, message):
        if self.callback:
            self.callback(message)

    def send_message(self, message):
        data = message.to_string()
        with self.sender_lock:
            logging.debug(f"TCP: sending data: {data}")
            self.client_socket.sendall(data)

    def close(self):
        if self.client_socket:
            self.client_socket.close()


class TCP:
    def __init__(self, address="127.0.0.1", port=7624):
        self.address = address
        self.port = port

    def connect(self, callback):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.address, self.port))
        handler = ConnectionHandler(sock, callback)

        handler_thread = threading.Thread(
            target=handler.wait_for_messages, daemon=True,
        )
        handler_thread.start()

        return handler



import datetime

from .defs import (DefBLOBVector, DefLightVector, DefNumberVector,
                   DefSwitchVector, DefTextVector, DefVector)
from .DelProperty import DelProperty
from .EnableBLOB import EnableBLOB
from .GetProperties import GetProperties
from .IndiMessage import IndiMessage
from .Message import Message
from .news import (NewBLOBVector, NewNumberVector, NewSwitchVector,
                   NewTextVector, NewVector)
from .sets import (SetBLOBVector, SetLightVector, SetNumberVector,
                   SetSwitchVector, SetTextVector, SetVector)


def now():
    return datetime.datetime.utcnow().isoformat()



import logging
from typing import Callable, Dict, Optional, Union

from indi import message
from indi.device.properties import Group as GroupDefinition
from indi.device.properties.definition.vectors import Vector
from indi.device.properties.instance.group import Group
from indi.message import IndiMessage
from indi.routing import Device, Router


class DriverMeta(type):
    def __new__(meta, name, bases, dct):
        for k in dct:
            if isinstance(dct[k], GroupDefinition):
                dct[k].property_name = k

        return super().__new__(meta, name, bases, dct)


class Driver(Device, metaclass=DriverMeta):
    onchange = None

    def __init__(self, name: Optional[str] = None, router: Optional[Router] = None):
        if name is not None:
            self._name = name
        elif self.__class__.name is not None:
            self._name = self.__class__.name

        self._groups: Dict[str, Group] = {
            k: Group(self, v) for k, v in self.__class__._group_definitions().items()
        }
        self._router = router

        self._vectors: Dict[str, Vector] = {}
        for k, group in self._groups.items():
            for k, vector in group.vectors.items():
                self._vectors[vector.name] = vector

        if self._router:
            self._router.register_device(self)

    @classmethod
    def _group_definitions(cls) -> Dict[str, GroupDefinition]:
        groups: Dict[str, GroupDefinition] = {}
        for base in cls.__bases__:
            if issubclass(base, Driver) or base is Driver:
                groups = {**groups, **base._group_definitions()}
        for k, v in cls.__dict__.items():
            if isinstance(v, GroupDefinition):
                groups[k] = v
        return groups

    @property
    def name(self) -> Optional[str]:
        return self._name

    def accepts(self, device: str) -> bool:
        return self.name == device

    def get_group(self, name: str) -> Optional[Group]:
        return self._groups.get(name)

    def send_message(self, msg: IndiMessage):
        if self._router and msg:
            self._router.process_message(msg, self)

    def message_from_client(self, msg: IndiMessage):
        if isinstance(msg, message.GetProperties):
            if not msg.name:
                for k, v in self._vectors.items():
                    self.send_message(v.to_def_message())
            else:
                v = self._vectors[msg.name]
                self.send_message(v.to_def_message())

        if isinstance(msg, message.news.NewVector):
            self._vectors[msg.name].from_new_message(msg)

    def trigger_callback(
        self, callback: Optional[Union[str, Callable]], sender, **kwargs
    ):
        if not callback:
            return

        if isinstance(callback, str):
            callback_fun = getattr(self, callback)
        else:
            callback_fun = callback

        try:
            callback_fun(sender, **kwargs)
        except Exception as e:
            logging.exception(e)



from typing import Optional

from indi.typing import Callback


class Group:
    def __init__(
        self, name: str, enabled=True, onchange: Optional[Callback] = None, vectors=None
    ):
        self.name = name
        self.vectors = vectors
        self.property_name = None
        self.onchange = onchange
        self.enabled = enabled

    def __get__(self, instance, objtype=None):
        return instance.get_group(self.property_name)

    def __getattr__(self, item):
        vector = self.vectors.get(item)
        if vector:
            return vector

        return self.__getattribute__(item)



from typing import Optional

from indi.device.properties.definition.group import Group as GroupDefinition
from indi.typing import Callback


class Group:
    def __init__(self, device, definition):
        self._device = device
        self._definition: GroupDefinition = definition
        self._vectors = {k: v.instance(self) for k, v in definition.vectors.items()}
        self._enabled = definition.enabled

    def __getattr__(self, item):
        vector = self._vectors.get(item)
        if vector:
            return vector

        return self.__getattribute__(item)

    @property
    def device(self):
        return self._device

    @property
    def name(self):
        return self._definition.name

    @property
    def vectors(self):
        return self._vectors

    @property
    def onchange(self) -> Optional[Callback]:
        return self._definition.onchange

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        self._enabled = value
        for k, v in self._vectors.items():
            self.device.send_message(v.to_def_message())
            self.device.send_message(v.to_set_message())



from typing import Type

from indi.device.properties.const import Permissions, State, SwitchRule
from indi.device.properties.definition.elements import (BLOB, Light, Number,
                                                        Switch, Text)
from indi.device.properties.instance import elements as instance_elements
from indi.device.properties.instance import vectors as instance_vectors
from indi.message import const


class Vector:
    element_class = Type[instance_elements.Element]
    instance_class: Type[instance_vectors.Vector]

    def __init__(
        self,
        name,
        label=None,
        state=State.OK,
        perm=Permissions.READ_WRITE,
        timeout=0,
        enabled=True,
        elements=None,
        onchange=None,
    ):
        self.name = name
        self.label = label or name
        self.state = state
        self.perm = perm
        self.timeout = timeout
        self.enabled = enabled
        self.onchange = onchange

        if not elements:
            raise Exception("No vector elements declared")

        for k, element in elements.items():
            if self.__class__.element_class and not isinstance(
                element, self.__class__.element_class
            ):
                raise Exception("")

        self.elements = elements

    def instance(self, group):
        return self.instance_class(group=group, definition=self)

    def __getattr__(self, item):
        element = self.elements.get(item)
        if element:
            return element

        return self.__getattribute__(item)


class NumberVector(Vector):
    element_class = Number
    instance_class = instance_vectors.NumberVector


class TextVector(Vector):
    element_class = Text
    instance_class = instance_vectors.TextVector


class SwitchVector(Vector):
    element_class = Switch
    instance_class = instance_vectors.SwitchVector

    RULES = SwitchRule

    def __init__(self, *args, rule=SwitchRule.ONE_OF_MANY, default_on=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.rule = rule
        if default_on is not None:
            default_on = (default_on,) if isinstance(default_on, str) else default_on
            for k, el in self.elements.items():
                if el.name in default_on:
                    el.default = const.SwitchState.ON


class LightVector(Vector):
    element_class = Light
    instance_class = instance_vectors.LightVector

    def __init__(self, *args, **kwargs):
        if kwargs.get("perm"):
            del kwargs["perm"]
        if kwargs.get("timeout"):
            del kwargs["timeout"]

        super().__init__(*args, **kwargs)


class BLOBVector(Vector):
    element_class = BLOB
    instance_class = instance_vectors.BLOBVector



from typing import Optional, Type, Union

from indi.message import checks
from indi.message.IndiMessage import IndiMessage
from indi.message.parts import OneBLOB, OneNumber, OneSwitch, OneText


class NewVector(IndiMessage):
    children_class: Union[Type[OneBLOB], Type[OneNumber], Type[OneSwitch], Type[OneText]]
    from_client = True

    def __init__(self, device, name, timestamp=None, children=None, **junk):
        self.device = device
        self.name = name
        self.timestamp = timestamp
        self.children = checks.children(children, self.children_class)


class NewBLOBVector(NewVector):
    children_class = OneBLOB


class NewNumberVector(NewVector):
    children_class = OneNumber


class NewSwitchVector(NewVector):
    children_class = OneSwitch


class NewTextVector(NewVector):
    children_class = OneText



import threading
import time
import uuid
from typing import Any

from indi import message
from indi.client.device import Device
from indi.client.elements import Element
from indi.client.vectors import Vector
from indi.message import IndiMessage, const


class Client:
    def __init__(self, control_connection, blob_connection):
        self.devices = {}
        self.control_connection = control_connection
        self.blob_connection = blob_connection
        self.control_connection_handler = None
        self.blob_connection_handler = None
        self.callbacks = []

    def __getitem__(self, key) -> Device:
        return self.devices[key]

    def __contains__(self, key) -> bool:
        return key in self.devices

    def get_device(self, name: str) -> Device:
        return self.devices.get(name)

    def set_device(self, name: str, device: Device):
        self.devices[name] = device
        self.control_connection_handler.send_message(
            message.EnableBLOB(device=name, value=const.BLOBEnable.NEVER)
        )
        self.blob_connection_handler.send_message(
            message.EnableBLOB(device=name, value=const.BLOBEnable.ONLY)
        )

    def process_message(self, msg: IndiMessage):
        device = None
        if isinstance(msg, message.DefVector):
            device = self.get_device(msg.device)
            if not device:
                device = Device(self, msg.device)
                self.set_device(msg.device, device)

        if isinstance(msg, message.SetVector):
            device = self.get_device(msg.device)

        if isinstance(msg, message.DelProperty):
            device = self.get_device(msg.device)

        if device:
            device.process_message(msg)

    def send_message(self, msg: IndiMessage):
        self.control_connection_handler.send_message(msg)

    def start(self):
        self.control_connection_handler = self.control_connection.connect(
            self.process_message
        )
        self.blob_connection_handler = self.blob_connection.connect(
            self.process_message
        )

        self.control_connection_handler.send_message(
            message.GetProperties(version="2.0")
        )

    def onupdate(
        self, *, callback, device=None, vector=None, element=None, what="value"
    ):
        uid = uuid.uuid4()
        callback_config = dict(
            what=what,
            device=device,
            vector=vector,
            element=element,
            callback=callback,
            uuid=uid,
        )

        self.callbacks.append(callback_config)
        return uid

    def rmonupdate(
        self,
        uuid=None,
        device=None,
        vector=None,
        element=None,
        what=None,
        callback=None,
    ):
        to_rm = list()
        for cb in self.callbacks:
            if (
                uuid in (None, cb["uuid"],)
                and device in (None, cb["device"],)
                and vector in (None, cb["vector"],)
                and element in (None, cb["element"],)
                and what in (None, cb["what"],)
                and callback in (None, cb["callback"])
            ):
                to_rm.append(cb)

        for cb in to_rm:
            cb["polling_enabled"] = False
            self.callbacks.remove(cb)

    def waitforupdate(
        self,
        device=None,
        vector=None,
        element=None,
        what=None,
        expect=None,
        initial=None,
        cmp=None,
        timeout=-1,
        polling_enabled=True,
        polling_delay=1.0,
        polling_interval=1.0,
    ):
        if cmp is None:
            cmp = lambda a, b: a == b

        lock = threading.Lock()

        def cb(sender, **kwargs):
            release = False

            if expect is None:
                if initial is None:
                    release = True
                elif initial != sender.value:
                    release = True
            else:
                try:
                    if what == "state" and cmp(sender.state, expect):
                        release = True
                except AttributeError:
                    pass

                try:
                    if what == "value" and cmp(sender.value, expect):
                        release = True
                except AttributeError:
                    pass

            if release and lock.locked():
                lock.release()

        if polling_enabled:

            def poll():
                kwargs = {}
                if device:
                    kwargs["device"] = device
                if vector:
                    kwargs["name"] = vector

                msg = message.GetProperties(version="2.0", **kwargs)

                time.sleep(polling_delay)
                while lock.locked():
                    self.send_message(msg)
                    time.sleep(polling_interval)

            t = threading.Thread(target=poll)
            t.start()

        lock.acquire()
        uid = self.onupdate(
            device=device, vector=vector, element=element, what=what, callback=cb
        )

        acquired = lock.acquire(timeout=timeout)

        self.rmonupdate(uuid=uid)

        if lock.locked():
            lock.release()

        if not acquired:
            raise Exception("Timeout occurred")

    def trigger_update(self, sender: Any[Element, Vector], what, **kwargs):
        sender_device = None
        sender_vector = None
        sender_element = None

        if isinstance(sender, Element):
            sender_device = sender.vector.device.name
            sender_vector = sender.vector.name
            sender_element = sender.name
        elif isinstance(sender, Vector):
            sender_device = sender.device.name
            sender_vector = sender.name

        for callback in self.callbacks:
            if (
                callback["device"] in (None, sender_device,)
                and callback["vector"] in (None, sender_vector,)
                and callback["element"] in (None, sender_element,)
                and callback["what"] in (None, what,)
            ):

                callback["callback"](sender, what=what, **kwargs)



from setuptools import find_packages, setup
from pathlib import Path


BUNDLES = (
    "websockets",
)

# -*- Installation Requires -*-


def strip_comments(l):
    return l.split('#', 1)[0].strip()


def _pip_requirement(req, *root):
    if req.startswith('-r '):
        _, path = req.split()
        return reqs(*root, *path.split('/'))
    return [req]


def _reqs(*f):
    path = (Path.cwd() / 'requirements').joinpath(*f)
    with path.open() as fh:
        reqs = [strip_comments(l) for l in fh.readlines()]
        return [_pip_requirement(r, *f[:-1]) for r in reqs if r]


def reqs(*f):
    return [req for subreq in _reqs(*f) for req in subreq]


def extras(*p):
    """Parse requirement in the requirements/extras/ directory."""
    return reqs('extras', *p)


def extras_require():
    """Get map of all extra requirements."""
    return {x: extras(x + '.txt') for x in BUNDLES}
    

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name="indipy",
    version="0.1.0",
    description="Python implementation of INDI server and client",
    url="http://github.com/wlatanowicz/indipy",
    author="Wiktor Latanowicz",
    author_email="indipy@wiktor.latanowicz.com",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["test*", "devices"]),
    zip_safe=False,
    install_requires=reqs('base.txt'),
    tests_require=reqs('tests.txt'),
    extras_require=extras_require(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import logging
import xml.etree.ElementTree as ET

from indi.message import IndiMessage


class Buffer:
    def __init__(self):
        self.data = ""

        self.allowed_tags = [m.tag_name() for m in IndiMessage.__all_subclasses__()]

    def append(self, data):
        self.data += data

    def _cleanup_buffer(self):
        start = len(self.data) - 1
        for tag in self.allowed_tags:
            start = min(start, self.data.find("<" + tag))

        if start >= 0:
            self.data = self.data[start:]

    def process(self, callback):
        self._cleanup_buffer()
        end = 0
        while len(self.data) > 0 and end >= 0:
            end = self.data.find(">", end)

            if end > 0:
                end += 1
                partial = self.data[0:end]

                try:
                    xml = ET.fromstring(partial)
                    self.data = self.data[end:]
                    end = 0
                    try:
                        message = IndiMessage.from_string(partial)
                        callback(message)
                    except Exception as ex:
                        logging.warning("Buffer: Contents is not a valid message")
                        logging.exception(ex)
                except ET.ParseError:
                    pass



import logging
import socket
import threading

from indi.routing import Client
from indi.transport import Buffer

logger = logging.getLogger(__name__)


class ConnectionHandler(Client):
    connections = []

    def __init__(self, client_socket, router):
        self.buffer = Buffer()
        self.client_socket = client_socket
        self.router = router
        self.sender_lock = threading.Lock()
        if self.router:
            self.router.register_client(self)

    @classmethod
    def handler(cls, router):
        def handler_func(client_socket):
            conn = cls(client_socket, router)
            cls.connections.append(conn)
            try:
                conn.wait_for_messages()
            except:
                pass

            conn.close()
            cls.connections.remove(conn)

        return handler_func

    def wait_for_messages(self):
        while True:
            logger.debug(f"TCP: waiting for data")
            message = self.client_socket.recv(1024)
            if not message:
                logger.debug(f"TCP: no data, breaking")
                break
            logger.debug(f"TCP: got data: {message}")
            self.buffer.append(message.decode("latin1"))
            self.buffer.process(self.message_from_client)

    def message_from_client(self, message):
        if self.router:
            self.router.process_message(message, sender=self)

    def message_from_device(self, message):
        data = message.to_string()
        with self.sender_lock:
            logger.debug(f"TCP: sending data: {data}")
            self.client_socket.sendall(data)

    def close(self):
        if self.client_socket:
            self.client_socket.close()
        if self.router:
            self.router.unregister_client(self)


class TCP:
    def __init__(self, address="0.0.0.0", port=7624, max_connections=5, router=None):
        self.address = address
        self.port = port
        self.max_connections = max_connections
        self.router = router

    def start(self):
        logger.info("Starting TCP INDI server on %s:%s", self.address, self.port)
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind((self.address, self.port,))
        server.listen(self.max_connections)

        try:
            while True:
                client_sock, address = server.accept()
                logger.info(
                    "Accepted connection from {}:{}".format(address[0], address[1])
                )
                client_handler = threading.Thread(
                    target=ConnectionHandler.handler(self.router),
                    args=(client_sock,),
                    daemon=True,
                )
                client_handler.start()
        except:
            for conn in ConnectionHandler.connections:
                conn.close()
            if server:
                server.close()



from indi.message import parts


class Element:
    def_message_class = None
    set_message_class = None
    new_message_class = None

    @classmethod
    def from_message(cls, vector, msg):
        for subclass in cls.__subclasses__():
            if isinstance(msg, subclass.def_message_class):
                return subclass(vector, msg)
        return None

    def __init__(self, vector, msg):
        self.name = msg.name
        self.label = msg.label
        self.vector = vector
        self._value = None
        self._new_value = None

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._new_value = value

    def process_message(self, msg):
        if isinstance(msg, self.set_message_class):
            old_value = self.value
            self._value = msg.value
            if self._value != old_value:
                self.vector.device.client.trigger_update(self, "value")

    def to_new_message(self):
        return self.new_message_class(name=self.name, value=self._new_value)


class Number(Element):
    def_message_class = parts.DefNumber
    set_message_class = parts.OneNumber
    new_message_class = parts.OneNumber


class Switch(Element):
    def_message_class = parts.DefSwitch
    set_message_class = parts.OneSwitch
    new_message_class = parts.OneSwitch


class Text(Element):
    def_message_class = parts.DefText
    set_message_class = parts.OneText
    new_message_class = parts.OneText


class Light(Element):
    def_message_class = parts.DefLight
    set_message_class = parts.OneLight
    new_message_class = parts.OneLight


class BLOB(Element):
    def_message_class = parts.DefBLOB
    set_message_class = parts.OneBLOB
    new_message_class = parts.OneBLOB



def children(value, child_class):
    if value is None:
        return []
    return value


def dictionary(value, dictionary_class):
    if value not in dictionary_class.__dict__.values():
        raise Exception(
            f'Invalid value: "{value}" not found in {dictionary_class.__name__}'
        )
    return value



from indi.message.IndiMessage import IndiMessage


class GetProperties(IndiMessage):
    from_device = True
    from_client = True

    def __init__(self, version, device=None, name=None, **junk):
        self.version = version
        self.device = device
        self.name = name



import xml.etree.cElementTree as ET

from indi.message.parts import IndiMessagePart


class IndiMessage:
    from_device = False
    from_client = False

    @classmethod
    def tag_name(cls):
        return cls.__name__[:1].lower() + cls.__name__[1:]

    @classmethod
    def __all_subclasses__(cls):
        subclasses = []
        for subclass in cls.__subclasses__():
            subclasses.append(subclass)
            for nested_subclass in subclass.__all_subclasses__():
                subclasses.append(nested_subclass)
        return subclasses

    @classmethod
    def from_xml(cls, xml):
        tag = xml.tag
        message_class = None

        for subclass in cls.__all_subclasses__():
            if subclass.tag_name() == tag:
                message_class = subclass

        if not message_class:
            raise Exception(f"Invalid message: {tag}")

        kwargs = xml.attrib

        children = []
        for child in xml:
            children.append(IndiMessagePart.from_xml(child))

        if len(children) > 0:
            kwargs["children"] = children

        if xml.text:
            kwargs["value"] = xml.text.strip()

        return message_class(**kwargs)

    @classmethod
    def from_string(cls, string):
        xml = ET.fromstring(string)
        return cls.from_xml(xml)

    def to_xml(self):

        kwargs = {
            k: str(v)
            for k, v in self.__dict__.items()
            if v is not None and k not in ("children",)
        }

        element = ET.Element(self.__class__.tag_name(), **kwargs)

        if hasattr(self, "children"):
            for child in self.children:
                child.to_xml(element)

        return element

    def to_string(self):
        xml = self.to_xml()
        return ET.tostring(xml)
